from typing import AsyncIterator, Callable, Sequence
from typing_extensions import TypeAlias

import aiohttp.web

from ponty.http.routes import mount_routes, route_iter


Provider: TypeAlias = Callable[[aiohttp.web.Application], AsyncIterator[None]]


def startmeup(
    *,
    port: int,
    providers: Sequence[Provider] = (),
    route_tables: Sequence[aiohttp.web.RouteTableDef] = (),
) -> None:
    """The main entrypoint for your application. Kicks off the server event loop.

    :param port: TCP/IP port for HTTP server. No default, must be provided
    :param Sequence[Provider] providers: see :ref:`Providers`
    :param route_tables: Optional sequence of
      `aiohttp.web.RouteTableDef <https://docs.aiohttp.org/en/stable/web_reference.html#routetabledef>`_
      instances

    """
    app = aiohttp.web.Application()

    for p in providers:
        app.cleanup_ctx.append(p)

    mount_routes(app, *route_tables)
    _echo_routes()

    aiohttp.web.run_app(app, port=port)


def _echo_routes():
    print()
    print("serving:")
    for method, path in route_iter():
        print(method, "\t", path)
    print()



from collections import deque
import typing
import warnings

from ponty.memo.cache.base import CacheStore, cache, cachemiss, Stampede
from ponty.memo.lock.local import locallock
from ponty.utils import now_millis


T = typing.TypeVar("T")


class _LRUStore(CacheStore[T]):

    def __init__(self, ttl_ms: int, maxsize: int):
        super().__init__()

        if ttl_ms < 0:
            raise ValueError("negative ttl is nonsense")
        self._ttl_ms: typing.Final[int] = ttl_ms

        if maxsize < 0:
            raise ValueError("negative maxsize is nonsense")
        if maxsize == 0:
            warnings.warn("Cache without maxsize may grow endlessly, beware.")
        self._maxsize: typing.Final[int] = maxsize

        self._cache: dict[str, tuple[typing.Union[int, None], T]] = dict()
        self._lru: deque[str] = deque()

    async def get(self, key: str) -> typing.Union[T, type[cachemiss]]:
        try:
            expiry, data = self._cache[key]
        except KeyError:
            return cachemiss

        if expiry is not None and expiry < now_millis():
            await self.remove(key)
            return cachemiss

        # rotate key to the end of the LRU queue
        self._lru.remove(key)
        self._lru.append(key)

        return data

    async def set(self, key: str, data: T) -> None:
        if self._ttl_ms == 0:
            expiry = None
        else:
            expiry = now_millis() + self._ttl_ms

        self._cache[key] = (expiry, data)
        self._lru.append(key)
        await self._sizecheck()

    async def remove(self, key: str) -> bool:
        try:
            self._cache.pop(key)
        except KeyError:
            return False
        else:
            self._lru.remove(key)
        return True

    async def _sizecheck(self) -> None:
        if not self._maxsize:
            return

        while len(self._lru) > self._maxsize:
            key = self._lru.popleft()
            self._cache.pop(key)


def localcache(
    *,
    ttl_ms: int = 0,
    maxwait_ms: int = 1000,
    pulse_ms: int = 50,
    maxsize: int = 128,
    name: str = "",
):
    """Process-RAM LRU memoizer with antistampede.

    Good for small frequently-used datasets (high ttl) or volatile
    stampede-likely objects (low ttl).

    :param ttl_ms: millis to expiry. Use 0 (the default) for no expiry
    :param maxwait_ms: millis to wait for a lock to resolve.
      Throws :class:`Stampede` error when :math:`(n * pulse\_ms) > maxwait\_ms`
    :param pulse_ms: antistampede recheck frequency
    :param maxsize: evict least recently used values once the cache reaches
      `maxsize` elements
    :param name: Optional. Providing a name registers the cache, so it can be
      used by :func:`invalidate`


    Here is an example that demonstrates stampede control.
    With `maxwait_ms = 0`, this operates as a mandatory lock:

    .. code-block:: python
        :emphasize-lines: 13

        import asyncio

        from ponty import get, render_json
        from ponty.memo import localcache


        @get("/cachetest")
        @render_json
        async def _locktest(_):
            return await _fetch()


        @localcache(maxwait_ms=0)
        async def _fetch():
            await asyncio.sleep(1)  # some expensive operation
            return {"key": "value"}


    For five simultaneous requests,
    the first acquires the lock and begins work
    while the others are immediately declined:

    .. code-block:: bash

        $ for i in {1..5}; do curl localhost:8080/cachetest -v & done

        < HTTP/1.1 409 Conflict
        < Date: Tue, 23 Aug 2022 21:49:12 GMT

        < HTTP/1.1 409 Conflict
        < Date: Tue, 23 Aug 2022 21:49:12 GMT

        < HTTP/1.1 409 Conflict
        < Date: Tue, 23 Aug 2022 21:49:12 GMT

        < HTTP/1.1 409 Conflict
        < Date: Tue, 23 Aug 2022 21:49:12 GMT

        < HTTP/1.1 200 OK
        < Date: Tue, 23 Aug 2022 21:49:13 GMT
        {"now": 1661291352844, "data": {"key": "value"}, "elapsed": 1006}


    Now that the item is cached,
    running this again gives five immediate cache hits:

    .. code-block:: bash

        $ for i in {1..5}; do curl localhost:8080/cachetest & done
        {"now": 1661291746863, "data": {"key": "value"}, "elapsed": 0}
        {"now": 1661291746863, "data": {"key": "value"}, "elapsed": 0}
        {"now": 1661291746863, "data": {"key": "value"}, "elapsed": 0}
        {"now": 1661291746863, "data": {"key": "value"}, "elapsed": 0}
        {"now": 1661291746863, "data": {"key": "value"}, "elapsed": 0}


    Bump to `maxwait_ms = 2000` and clear the cache:

    .. code-block:: bash

        $ for i in {1..5}; do curl localhost:8080/cachetest & done
        {"now": 1661396864744, "data": {"key": "value"}, "elapsed": 1003}
        {"now": 1661396864744, "data": {"key": "value"}, "elapsed": 1043}
        {"now": 1661396864744, "data": {"key": "value"}, "elapsed": 1044}
        {"now": 1661396864744, "data": {"key": "value"}, "elapsed": 1045}
        {"now": 1661396864744, "data": {"key": "value"}, "elapsed": 1045}

    The first request is the only one doing work here;
    the others wait up to two seconds,
    checking in every `pulse_ms = 50` milliseconds (the default)
    to see if a result is available.

    """
    antistampede = locallock(
        maxwait_ms=maxwait_ms,
        pulse_ms=pulse_ms,
        timeout_error=Stampede,
    )

    return cache(
        store=_LRUStore(ttl_ms, maxsize),
        antistampede=antistampede,
        name=name,
    )



import abc
from contextlib import asynccontextmanager
import functools
import hashlib
import json
import typing
import warnings

from ponty.memo.lock.base import Lock, Locked
from ponty.registry import Registry


class Stampede(Locked):
    """Raised when the cache comes under high load and the lock times out.

    Inherits :class:`Locked <ponty.memo.Locked>`.

    """


class cachemiss:
    """Sentinel value representing a cache miss.

    Returned by subclasses of :class:`CacheStore`.

    """




T = typing.TypeVar("T")


class CacheStore(abc.ABC, typing.Generic[T]):
    """Abstract Base Class. Container for cached items.

    :func:`cache` expects two abstract methods,
    :meth:`get <ponty.memo.CacheStore.get>` and
    :meth:`set <ponty.memo.CacheStore.set>`, and
    :func:`invalidate` expects one,
    :meth:`remove <ponty.memo.CacheStore.remove>`.
    All three must return awaitables.

    Generic on one variable, the type T of the cached item.

    """

    @abc.abstractmethod
    async def get(self, key: str) -> typing.Union[T, type[cachemiss]]:
        """Fetch the cached value.

        Errors are not captured.

        :param key: unique id for the cached item
        :returns: the cached item,
          or the :class:`cachemiss` sentinel if it does not exist
        :rtype: Union[T, type[:class:`cachemiss`]]

        """

    @abc.abstractmethod
    async def set(self, key: str, data: T) -> None:
        """Add an item to the cache.

        :param key: unique id for the item
        :param data: the item to be cached

        """

    @abc.abstractmethod
    async def remove(self, key: str) -> bool:
        """Remove an item from the cache.

        :param key: unique id for the cached item
        :returns: True if the item is found and removed, False otherwise

        """




_registry = Registry[tuple[CacheStore, Lock]]()


def cache(store: CacheStore, antistampede: Lock, name: str = ""):
    """Cache the decorated function's return value. Not thread-safe.

    As a rule of thumb, storage mechanics for a cache are
    implemented in `store`, and synchronization rules are handled
    by `antistampede`.
    See the :func:`localcache`
    `source ↗ <https://github.com/csira/ponty/blob/main/ponty/memo/cache/local.py>`__
    for an example.

    :param store: container, implementing get/set mechanics for cached items
    :param antistampede: instance of :class:`ponty.memo.Lock`;
      blocks on simultaneous access to the requested key to avoid a stampede
    :param name: if provided, registers the cache for use by
      :func:`invalidate`. Must be unique

    """
    def wraps(f):
        if name:
            _registry.add(name, (store, antistampede))

        @functools.wraps(f)
        async def wrapper(*a, **kw):
            key = _hash_args(*a, **kw)

            async with antistampede.lock(key):
                data = await store.get(key)
                if data is not cachemiss:
                    return data

                data = await f(*a, **kw)
                await store.set(key, data)

            return data

        return wrapper
    return wraps


@asynccontextmanager
async def invalidate(cachename: str, *a, **kw) -> typing.AsyncIterator[None]:
    """Context manager. Invalidates an item in the cache.

    Waits for then holds the cache's lock until the context manager exits.
    This prevents simultaneous fetches from re-cache-ing the item before
    mutations are committed. This also means functions calling 'invalidate'
    CANNOT be wrapped with a cache decorator, as the decorator will already
    hold the lock.

    :param cachename: name of the registered cache.
      Must match :func:`cache`'s `name` parameter

    Note the \*arg/\**kwarg combo must match the arguments to the
    cache-decorated function EXACTLY in order to get a key hit. E.g.,

    .. code-block:: python
        :emphasize-lines: 2,7,9

        @localcache(name="foo")
        async def get_foo(foo_id):
            ...


        async def update_foo(foo_id, ...):
            async with invalidate("foo", foo_id):  # this hits
                ...
            async with invalidate("foo", foo_id=foo_id):  # this does not
                ...

    """
    key = _hash_args(*a, **kw)

    try:
        store, antistampede = _registry.get(cachename)
    except KeyError:
        warnings.warn(f"'{cachename}' is not registered", category=RuntimeWarning)
        yield
    else:
        async with antistampede.lock(key):
            await store.remove(key)
            yield


def _hash_args(*a, **kw) -> str:
    args = a + tuple(sorted(kw.items()))
    encoded = json.dumps(args).encode("utf-8")
    return hashlib.sha1(encoded).hexdigest()



__all__ = [
    "dataclass_to_jsonschema",
    "Annotation",
    "validator_for_dataclass",
    "validator_for_schema",
    "validator_from_file",
]


import functools
import json

from jsonschema.validators import validator_for  # type: ignore

from ponty.http.schema.build import dataclass_to_jsonschema, Annotation


def validator_for_schema(schema):
    validator = validator_for(schema)
    validator.check_schema(schema)
    return validator(schema)


@functools.cache
def validator_for_dataclass(cls):
    schema = dataclass_to_jsonschema(cls)
    return validator_for_schema(schema)


@functools.cache
def validator_from_file(filepath: str):
    with open(filepath) as f:
        schema = json.load(f)
    return validator_for_schema(schema)



import functools
import typing

from aiohttp.hdrs import (
    METH_DELETE,
    METH_GET,
    METH_PATCH,
    METH_POST,
    METH_PUT,
)
import aiohttp.web

from ponty.errors import error_trap


_route_table = aiohttp.web.RouteTableDef()


def route(method: str, path: str, **kw):
    """
    Bind a (method, path) pair to its handler, and register with the master route table.

    :class:`PontyError`'s are automatically trapped and handled here.

    :param method: HTTP method, e.g. GET
    :param path: Relative path to resource.
      May contain braces {} for dynamic components,
      or references to :class:`RouteParameter` instances in an f-string
    :param kw: Additional arguments discussed
      `here <https://docs.aiohttp.org/en/latest/web_reference.html#aiohttp.web.UrlDispatcher.add_get>`__

    """
    def wraps(f):
        @_route_table.route(method, path, **kw)
        @error_trap
        @functools.wraps(f)
        async def wrapper(*a, **kw):
            return await f(*a, **kw)
        return wrapper
    return wraps


delete = functools.partial(route, METH_DELETE)
get = functools.partial(route, METH_GET)
patch = functools.partial(route, METH_PATCH)
post = functools.partial(route, METH_POST)
put = functools.partial(route, METH_PUT)




def route_iter() -> typing.Iterator[tuple[str, str]]:
    """Iterator, over routes on the primary route table.

    :return: (method, path) pairs
    :rtype: Iterator[tuple[str, str]]

    """
    for route in ( typing.cast(aiohttp.web.RouteDef, r) for r in _route_table ):
        yield route.method, route.path


def mount_routes(
    app: aiohttp.web.Application,
    *route_tables: aiohttp.web.RouteTableDef,
) -> None:

    app.router.add_routes(_route_table)
    for rt in route_tables:
        app.router.add_routes(rt)



import aiohttp.web

from ponty.errors import raise_status
from ponty.http.expect.body import (
    ValidatedJsonBody,
    ParsedJsonBody,
)
from ponty.http.expect.header import (
    Header,
    ContentLength,
    ContentType,
)
from ponty.http.expect.query import QueryParameter, QueryParameterEnum
from ponty.http.expect.req import (
    expect,
    Request,
    JsonBody,
    TextBody,
)
from ponty.http.expect.route import (
    RouteParameter,
    PosIntRouteParameter,
    StringRouteParameter,
)


class AIOHttpReq:
    """Forwards along the
    `aiohttp.web.Request <https://docs.aiohttp.org/en/latest/web_reference.html#aiohttp.web.Request>`__.

    """
    def __get__(self, obj: Request, objtype: type[Request] = None) -> aiohttp.web.Request:
        return obj.req


class Cookie:
    """Extracts request cookie `name`.

    :param str name: cookie name

    :param bool required:
      if True, throws a 400 if the cookie is not provided

    :param str default:
      default value, if the cookie is not present

    :param int errorcode:
      HTTP status code raised, if the cookie is required and not supplied

    """
    def __init__(
        self,
        *,
        name: str,
        required: bool = False,
        default: str = "",
        errorcode: int = 400,
    ):
        self._name = name
        self._required = required
        self._default = default
        self._errorcode = errorcode

    def __get__(self, obj: Request, objtype: type[Request] = None) -> str:
        try:
            return obj.req.cookies[self._name]
        except KeyError:
            if self._required:
                raise_status(self._errorcode)
            return self._default



import typing


T = typing.TypeVar("T")


class Registry(typing.Generic[T]):

    def __init__(self):
        self._registry: dict[str, T] = {}

    def __contains__(self, name: str) -> bool:
        return name in self._registry

    def add(self, name: str, obj: T) -> None:
        if name in self:
            raise RuntimeError(f"'{name}' already registered")
        self._registry[name] = obj

    def get(self, name: str) -> T:
        return self._registry[name]



import asyncio
from contextlib import asynccontextmanager
import typing

import aiohttp

from ponty.http.go import Provider
from ponty.registry import Registry


_registry = Registry[tuple[asyncio.Semaphore, aiohttp.ClientSession]]()
_defaultname: typing.Final[str] = "_default"


def http_client_provider(
    *,
    name: str = _defaultname,
    timeout: int = 5,
    concurrency: int = 100,
    headers: dict[str, typing.Any] = None,
    cookies: dict[str, typing.Any] = None,
    **kw
) -> Provider:
    """Builds an HTTP Client Session provider.

    Wraps `aiohttp.ClientSession <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession>`_.

    :param name: unique name, used by :func:`lease_http_client`
    :param timeout: total timeout for the complete operation, in seconds
    :param concurrency: maximum number of concurrent requests
    :param headers: headers to send with each request
    :param cookies: cookies to send with each request
    :param kw: additional parameters, specified `here <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession>`__
    :return: a provider, which should be passed to :func:`startmeup`


    .. code-block:: python

        startmeup(
            port=8080,
            providers=(
                http_client_provider(
                    timeout=30,
                    concurrency=10,
                ),
            ),
        )

    """
    async def provider(_) -> typing.AsyncIterator[None]:
        sem = asyncio.Semaphore(concurrency)
        sess = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=timeout),
            headers=headers,
            cookies=cookies,
            **kw
        )

        _registry.add(name, (sem, sess))

        try:
            yield
        finally:
            await sess.close()
            await asyncio.sleep(0.5)  # sleep a beat to allow underlying connections to close

    return provider


@asynccontextmanager
async def lease_http_client(name: str = _defaultname) -> typing.AsyncIterator[aiohttp.ClientSession]:
    """
    Context manager. Fetch the named http client session, leasing one
    of its request positions or blocking until one becomes available,
    then return it to the pool after the block is exited.

    :param name: see :func:`http_client_provider` parameter ``name``
    :raises KeyError: if the named session was not registered (via :func:`startmeup`)
    :yields: `aiohttp.ClientSession <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession>`__


    .. code-block:: python

        async with lease_http_client() as sess:
            async with sess.get("https://www.python.org/") as resp:
                html = await resp.text()

    """
    sem, sess = _registry.get(name)
    async with sem:
        yield sess



__all__ = [
    "startmeup", "Provider",
    "get", "post", "put", "delete", "patch", "route", "route_iter",
    "expect",
    "JsonBody", "TextBody", "ValidatedJsonBody", "ParsedJsonBody",
    "Header", "ContentLength", "ContentType",
    "QueryParameter", "QueryParameterEnum",
    "Request",
    "RouteParameter", "PosIntRouteParameter", "StringRouteParameter",
    "AIOHttpReq",
    "Cookie",
    "render_json",
    "http_client_provider", "lease_http_client",
    "Annotation", "dataclass_to_jsonschema",
]


from ponty.http.client import http_client_provider, lease_http_client
from ponty.http.expect import (
    expect,
    JsonBody, TextBody, ValidatedJsonBody, ParsedJsonBody,
    Header, ContentLength, ContentType,
    QueryParameter, QueryParameterEnum,
    Request,
    RouteParameter, PosIntRouteParameter, StringRouteParameter,
    AIOHttpReq,
    Cookie,
)
from ponty.http.go import startmeup, Provider
from ponty.http.render import render_json
from ponty.http.routes import get, post, put, delete, patch, route, route_iter
from ponty.http.schema import Annotation, dataclass_to_jsonschema



import aiohttp.web


class Header:
    """
    Extracts request header `key`.

    :param str key: header name

    :param bool required:
      if True, throws a 400 if the request header is not provided

    :param str default:
      default value, if the header does not appear on the request

    """
    def __init__(self, *, key: str, required: bool = False, default: str = ""):
        self._key = key
        self._required = required
        self._default = default

    def __get__(self, obj, objtype=None) -> str:
        try:
            return obj.req.headers[self._key]
        except KeyError:
            if self._required:
                raise aiohttp.web.HTTPBadRequest
            return self._default


class ContentLength(Header):
    """Inherits :class:`Header`. Extracts the "content-length" header."""

    def __init__(self, **kw):
        super().__init__(key="content-length", **kw)


class ContentType(Header):
    """Inherits :class:`Header`. Extracts the "content-type" header."""

    def __init__(self, **kw):
        super().__init__(key="content-type", **kw)



import functools

import aiohttp.web

from ponty.utils import now_millis


def render_json(f):
    """
    Encode the decorated function's return value with
    `json.dumps <https://docs.python.org/3/library/json.html#json.dumps>`__,
    and package it into an HTTP
    `response <https://docs.aiohttp.org/en/stable/web_reference.html#json-response>`__.
    The JSON has the form:

    .. code-block:: text

        {
            "data": <function return value>,
            "elapsed": <time elapsed, in millis>,
            "now": <current epoch timestamp, in millis>,
        }

    e.g.,

    .. code-block:: python
        :emphasize-lines: 7

        from ponty import get, render_json


        @get("/hello")
        @render_json
        async def handler(_):
            return {"greeting": "hello world!"}


    .. code-block:: bash
        :emphasize-lines: 4

        $ curl localhost:8080/hello | python -m json.tool
        {
            "data": {
                "greeting": "hello world!"
            },
            "elapsed": 1,
            "now": 1660440618107
        }

    """
    @functools.wraps(f)
    async def wrapper(*a, **kw) -> aiohttp.web.Response:
        return aiohttp.web.json_response({
            "now": (start := now_millis()),
            "data": await f(*a, **kw),
            "elapsed": now_millis() - start,
        })
    return wrapper


def boomerang_cookie(cookie_name: str):
    def wraps(f):
        @functools.wraps(f)
        async def wrapper(req: aiohttp.web.Request) -> aiohttp.web.Response:
            cookie_val = req.cookies[cookie_name]
            resp = await f(req)
            resp.cookies[cookie_name] = cookie_val
            return resp
        return wrapper
    return wraps



import functools
import json
import typing
from typing import Awaitable, Callable, TypeVar
from typing_extensions import ParamSpec

import aiohttp.web


class PontyError(Exception):
    """
    PontyError's may be raised anywhere within your program to immediately
    respond with an HTTP status code. Unhandled errors are reraised as
    `aiohttp.web.HTTPException <https://docs.aiohttp.org/en/stable/web_reference.html#aiohttp.web.HTTPException>`__.

    :param text: response body, as text/plain
    :param body: json-serializable response body
    :param kw: see parameter description for
      `aiohttp.web.HTTPException <https://docs.aiohttp.org/en/stable/web_reference.html#aiohttp.web.HTTPException>`__

    """
    status_code: int = 500  #: HTTP status code for the response. See the official IANA registry `here <https://www.iana.org/assignments/http-status-codes/http-status-codes.xhtml>`__

    def __init__(
        self,
        *,
        text: str = None,
        body: typing.Any = None,
        **kw
    ):
        if text is not None and body is not None:
            raise ValueError

        super().__init__()
        self.text = text
        self.body = body
        self.kw = kw


class DoesNotExist(PontyError):
    """Returns a 404. Inherits :class:`PontyError`."""

    status_code = 404


class ValidationError(PontyError):
    """Returns a 400. Inherits :class:`PontyError`."""

    status_code = 400




P = ParamSpec("P")
R = TypeVar("R")


def error_trap(f: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
    @functools.wraps(f)
    async def wrapper(*a: P.args, **kw: P.kwargs) -> R:
        try:
            return await f(*a, **kw)
        except PontyError as e:
            raise_status(
                e.status_code,
                text=e.text,
                body=e.body,
                **e.kw
            )
    return wrapper


def raise_status(
    status: int,
    *,
    text: str = None,
    body: typing.Any = None,
    content_type: str = None,
    **kw
) -> typing.NoReturn:
    """
    Raise an `aiohttp.web.HTTPException <https://docs.aiohttp.org/en/stable/web_reference.html#aiohttp.web.HTTPException>`_
    for the given status code.

    :param status: HTTP status code
    :param text: response body
    :param body: json-serializable response body.
      Automatically sets the content-type header to 'application/json'.
      Only one of `text` or `body` may be supplied.
    :param content_type: value for the response's content-type header
    :param kw: additional params for `aiohttp.web.HTTPException <https://docs.aiohttp.org/en/stable/web_reference.html#aiohttp.web.HTTPException>`_


    .. code-block:: python
        :emphasize-lines: 3

        @get("/fail")
        async def handler(_):
            raise_status(400, body={"a": 1})


    .. code-block:: bash
        :emphasize-lines: 7,13

        $ curl localhost:8080/fail -v
        > GET /fail HTTP/1.1
        > Host: localhost:8080
        > User-Agent: curl/7.64.1
        > Accept: */*
        >
        < HTTP/1.1 400 Bad Request
        < Content-Type: application/json; charset=utf-8
        < Content-Length: 8
        < Date: Wed, 10 Aug 2022 02:28:55 GMT
        < Server: Python/3.9 aiohttp/3.7.3
        <
        {"a": 1}

    """
    if text is not None and body is not None:
        raise ValueError

    if body:
        text = json.dumps(body)
        content_type = "application/json"

    class MyExc(aiohttp.web.HTTPException):
        status_code = status

    raise MyExc(text=text, content_type=content_type, **kw)



from ponty.memo.cache.base import CacheStore, cache, invalidate, cachemiss, Stampede
from ponty.memo.cache.local import localcache
from ponty.memo.lock.base import SentinelStore, Lock, Locked
from ponty.memo.lock.local import locallock


__all__ = [
    "CacheStore", "cache", "invalidate", "cachemiss", "Stampede", "localcache",
    "SentinelStore", "Lock", "Locked", "locallock",
]



import asyncio
import functools
import random
import time


def now_millis() -> int:
    return int(time.time() * 1000)


def logical_xor(m, n) -> bool:
    """Logical xor. True if exactly one of {m, n} is truth-y.

    Equivalent to
        (m and not n) or (not m and n)

    NB: xor on two bools is logical since, on integers, logical xor is identical
    to bitwise xor when the domain is limited to {0, 1}

    """
    return bool(m) ^ bool(n)




def retry(
    *excs: type[Exception],
    max_retries: int = 1,
    t1_ms: int = 100,
    t2_ms: int = 4000,
    backoff_factor: float = 1,
    anticollision: bool = False,
):
    """Retry the wrapped function if the specified exception(s) are raised.

    Supports sync & async functions.

    :param excs: Exception class(es) to trap
    :param max_retries: max number of retries to attempt
    :param t1_ms: initial interval between retries, in millis
    :param t2_ms: max interval between retries, in millis
    :param backoff_factor:
      rate reduction factor for exponential backoff :math:`t = b^c`, where

        | t is the delay factor between calls
        | b = `backoff_factor` is the base
        | c is the number of failures observed so far

      :math:`b > 1` will reduce the retry rate,
      :math:`b = 1` will give consistent retry intervals, and
      :math:`b < 1` will accelerate retries (unusual choice).
      Set to 2 for run-of-the-mill binary exponential backoff.
    :param anticollision: a deterministic algorithm may be unsuitable when
      errors are caused by collisions, because each client will sleep for the
      same amount of time, leading to subsequent collisions ad infinitum.
      Pass True to treat :math:`b^c` as an upper bound on the
      time delay; in this case we will sleep for :math:`k * t1\_ms`, where k is
      a random integer on :math:`[0, \\lfloor b^c  \\rfloor)`.


    .. code-block:: python

        import random
        import time

        from ponty import retry


        @retry(KeyError, ValueError, max_retries=10, backoff_factor=2)
        def getval():
            i = random.choice(range(5))
            print(f"i = {i}, now = {time.time()}")

            if i == 0:
                return 42
            if i % 2:
                raise KeyError
            raise ValueError


    .. code-block:: bash

        >>> print(getval())
        i = 3, now = 1660102025.223419
        i = 3, now = 1660102025.328576
        i = 4, now = 1660102025.533844
        i = 1, now = 1660102025.939289
        i = 1, now = 1660102026.744708
        i = 2, now = 1660102028.347733
        i = 0, now = 1660102031.549533
        42

    """
    def backoff(n_failures: int) -> float:
        k = backoff_factor ** n_failures
        if anticollision:
            k = random.choice(range(int(k)))
        return min(k * t1_ms, t2_ms) * 0.001

    def wraps(f):

        @functools.wraps(f)
        def sync_wrapper(*a, **kw):
            for n_failures in range(max_retries+1):
                try:
                    return f(*a, **kw)
                except excs as e:
                    if n_failures == max_retries:
                        raise e

                time.sleep(backoff(n_failures))

        @functools.wraps(f)
        async def async_wrapper(*a, **kw):
            for n_failures in range(max_retries+1):
                try:
                    return await f(*a, **kw)
                except excs as e:
                    if n_failures == max_retries:
                        raise e

                await asyncio.sleep(backoff(n_failures))

        if asyncio.iscoroutinefunction(f):
            return async_wrapper
        return sync_wrapper

    return wraps



from dataclasses import dataclass
import sys
from typing import Any, Protocol


if sys.version_info >= (3, 10):
    @dataclass
    class DataclassProtocol(Protocol):
        ...

else:
    class DataclassProtocol(Protocol):
        __dataclass_fields__: dict[str, Any]



import typing

from ponty.http.expect.req import Request


T = typing.TypeVar("T")


class RouteParameter(typing.Generic[T]):
    """
    The *RouteParameter* descriptor is used to identify and extract
    variable resources from the URI. It is generic on one variable.

    When invoked from an instance, returns the variable part of the resolved route.

    When invoked from a class, returns the `{identifier:regex}` rendering
    required for path matching.
    (See `Variable Resources <https://docs.aiohttp.org/en/stable/web_quickstart.html#variable-resources>`_.)
    Particularly useful in an f-string (example below).

    :param str pattern: custom regular expression to match the variable part
    :param cast_to: converts the variable part to type `T`


    New reusable "matchers" can be created like so:

    .. code-block:: python
       :emphasize-lines: 9,17,22

        from ponty import (
            expect,
            get,
            Request,
            RouteParameter,
        )


        class FiveLetters(RouteParameter[str]):

            def __init__(self):
                super().__init__(pattern="[a-zA-Z]{5}")


        class MyReq(Request):

            id = FiveLetters()


        @get(f"/obj/{MyReq.id}")
        @expect(MyReq)
        async def handler(id: str):
            ...

    """
    def __init__(
        self,
        *,
        pattern: str,
        cast_to: typing.Callable[[str], T] = None,
    ):
        self._pattern = pattern
        self._cast_func = cast_to

    def __set_name__(self, owner: type[Request], name: str):
        self._key = name

    @typing.overload
    def __get__(self, obj: None, objtype: type[Request]) -> str: ...

    @typing.overload
    def __get__(self, obj: Request, objtype: type[Request]) -> T: ...

    def __get__(
        self,
        obj: typing.Optional[Request],
        objtype: type[Request] = None,
    ) -> typing.Union[T, str]:

        if obj is None:
            return f"{{{self._key}:{self._pattern}}}"

        val = obj.req.match_info[self._key]
        if self._cast_func:
            return self._cast_func(val)
        return val


class PosIntRouteParameter(RouteParameter[int]):
    """
    Inherits :class:`RouteParameter`.
    Matches on a sequence of digits, ``\d+``.

    """
    def __init__(self):
        super().__init__(pattern="\d+", cast_to=int)


class StringRouteParameter(RouteParameter[str]):
    """
    Inherits :class:`RouteParameter`.
    Matches on strings, ``\w+``.

    """
    def __init__(self):
        super().__init__(pattern="\w+")



import json
import typing

import aiohttp.web
import jsonschema  # type: ignore

from ponty.http.schema import (
    dataclass_to_jsonschema,
    validator_for_schema,
    validator_from_file,
)
from ponty.http.expect.req import JsonBody, Request
from ponty.types_ import DataclassProtocol
from ponty.utils import logical_xor


class ValidatedJsonBody(JsonBody):
    """
    Validates the deserialized request body against the given jsonschema.
    Invalid input immediately triggers a 400 response.

    Inherits :class:`JsonBody`.

    :param filepath: jsonschema file location
      (absolute, relative to PYTHONPATH, or relative to working directory).


    .. code-block:: json
        :caption: validator.json

        {
          "$schema": "http://json-schema.org/draft-07/schema#",
          "type": "object",
          "properties": {
            "first_name": { "type": "string" },
            "last_name": { "type": "string" }
          },
          "required": ["first_name", "last_name"]
        }


    .. code-block:: python

        from ponty import (
            expect,
            post,
            render_json,
            Request,
            ValidatedJsonBody,
        )


        class MyReq(Request):

            body = ValidatedJsonBody(filepath="validator.json")


        @post("/test")
        @expect(MyReq)
        @render_json
        async def handler(body):
            return {"name": f"{body['last_name']}, {body['first_name']}"}


    .. code-block:: bash
        :caption: success

        $ curl localhost:8080/test \\
            -H "content-type:application/json" \\
            -d '{"first_name": "Donald", "last_name": "Duck"}' | python -m json.tool
        {
            "data": {
                "name": "Duck, Donald",
            },
            "elapsed": 0,
            "now": 1660505985510
        }


    .. code-block:: bash
        :caption: error, "last_name" omitted
        :emphasize-lines: 6,9

        $ curl localhost:8080/test \\
            -H "content-type:application/json" \\
            -d '{"first_name": "Donald"}' \\
            -v
        ...
        < HTTP/1.1 400 Bad Request
        < Content-Type: text/plain; charset=utf-8
        ...
        'last_name' is a required property

    """
    def __init__(
        self,
        *,
        filepath: str = None,
        schema: dict[str, typing.Any] = None,
    ):
        super().__init__()

        if not logical_xor(filepath, schema):
            raise RuntimeError("only one of filepath, schema may be specified")

        if filepath:
            self._validator = validator_from_file(filepath)
        else:
            self._validator = validator_for_schema(schema)

    def __get__(self, obj: Request, objtype: type[Request]):
        body = super().__get__(obj, objtype)

        try:
            self._validator.validate(body)
        except jsonschema.ValidationError as e:
            raise aiohttp.web.HTTPBadRequest(text=e.message)

        return body


D = typing.TypeVar("D", bound=DataclassProtocol)


class ParsedJsonBody(ValidatedJsonBody, typing.Generic[D]):
    """Holds the validated request body, structured as an instance of `cls`.

    Inherits :class:`ValidatedJsonBody`.

    :param dataclass cls: dataclass, into which the request body is marshalled.
      A custom jsonschema validator is automatically created by
      :func:`ponty.dataclass_to_jsonschema`

    :param filepath: if provided, path to the jsonschema file
      (see :class:`ValidatedJsonBody`).
      Supplants the dataclass-built default


    .. code-block:: python

        from dataclasses import asdict, dataclass
        import typing

        from ponty import (
            expect,
            post,
            render_json,
            Annotation,
            ParsedJsonBody,
            Request,
        )


        @dataclass(frozen=True)
        class Person:

            first_name: str
            last_name: str
            favorite_color: typing.Optional[str]
            height: typing.Annotated[int, Annotation(description="cm")]

            class JsonSchema:
                annotation = Annotation(description="request shape for a person")


        class NewPersonReq(Request):

            person = ParsedJsonBody(Person)


        @post("/person")
        @expect(NewPersonReq, mimetype="application/json")
        @render_json
        async def handler(person: Person):
            return asdict(person)


    .. code-block:: json
        :caption: the generated jsonschema
        :emphasize-lines: 22,27

        {
          "$schema": "http://json-schema.org/draft-07/schema#",
          "properties": {
            "first_name": {
              "type": "string"
            },
            "last_name": {
              "type": "string"
            },
            "favorite_color": {
              "anyOf": [
                {
                  "type": "string"
                },
                {
                  "type": "null"
                }
              ]
            },
            "height": {
              "type": "integer",
              "description": "cm"
            }
          },
          "title": "Person",
          "type": "object",
          "description": "request shape for a person",
          "required": [
            "first_name",
            "last_name",
            "favorite_color",
            "height"
          ]
        }

    (note the highlighted lines, controlled by :class:`Annotation`)


    .. code-block:: bash
        :caption: success, echo back the request body

        $ curl localhost:8080/person \\
            -H "content-type:application/json" \\
            -d '{"first_name": "Mickey", "last_name": "Mouse", "favorite_color": "blue", "height": 1000}' \\
            -v | python -m json.tool
        > POST /person HTTP/1.1
        > Host: localhost:8080
        > User-Agent: curl/7.64.1
        > Accept: */*
        > content-type:application/json
        > Content-Length: 83
        >
        < HTTP/1.1 200 OK
        < Content-Type: application/json; charset=utf-8
        < Content-Length: 129
        < Date: Sun, 14 Aug 2022 21:04:35 GMT
        < Server: Python/3.9 aiohttp/3.7.3
        <
        {
            "data": {
                "favorite_color": "blue",
                "first_name": "Mickey",
                "height": 1000,
                "last_name": "Mouse"
            },
            "elapsed": 0,
            "now": 1660511075634
        }


    .. code-block:: bash
        :caption: error, missing "height"
        :emphasize-lines: 6,9

        $ curl localhost:8080/person \\
            -H "content-type:application/json" \\
            -d '{"first_name": "Mickey", "last_name": "Mouse", "favorite_color": "blue"}' \\
            -v
        ...
        < HTTP/1.1 400 Bad Request
        < Content-Type: text/plain; charset=utf-8
        ...
        'height' is a required property


    .. code-block:: bash
        :caption: error, wrong type for "height"
        :emphasize-lines: 6,8

        $ curl localhost:8080/person \\
            -H "content-type:application/json" \\
            -d '{"first_name": "Mickey", "last_name": "Mouse", "favorite_color": "blue", "height": "abc"}' \\
            -v
        ...
        < HTTP/1.1 400 Bad Request
        ...
        'abc' is not of type 'integer'

    """
    def __init__(self, cls: type[D], *, filepath: str = None):
        self._cls = cls

        if filepath:
            super().__init__(filepath=filepath)
        else:
            super().__init__(schema=dataclass_to_jsonschema(cls))

    def __get__(self, obj: Request, objtype: type[Request]) -> D:
        body = super().__get__(obj, objtype)
        return self._cls(**body)



from ponty.memo.lock.base import SentinelStore, Lock


class _Sentinels(SentinelStore):

    def __init__(self):
        super().__init__()
        self._sentinels: set[str] = set()

    async def exists(self, key: str) -> bool:
        return key in self._sentinels

    async def add(self, key: str) -> None:
        self._sentinels.add(key)

    async def remove(self, key: str) -> None:
        self._sentinels.remove(key)


def locallock(**kw) -> Lock:
    """Mutex; coordinates across coroutines. Not thread-safe.

    :param maxwait_ms: total milliseconds to wait, before raising
      an exception with the caller.
      Raises `timeout_error` when :math:`n * pulse\_ms > maxwait\_ms`.
      If the lock is already held, the default `0` will cause the
      `timeout_error` to be raised immediately on a simultaneous access attempt

    :param pulse_ms: milliseconds between subsequent attempts
      to acquire the lock

    :param timeout_error: exception class raised when `maxwait_ms` exceeded.
      If provided, must be a subclass of :class:`Locked <ponty.memo.Locked>`


    A simple example that sleeps whenever the lock is acquired,
    in order to run out the 5-second clock on other waiting requests:

    .. code-block:: python
        :emphasize-lines: 9,19

        import asyncio
        import random
        import time

        from ponty import get, render_json
        from ponty.memo import locallock, Locked


        mylock = locallock(maxwait_ms=5000)  # concurrent requests wait 5 seconds max


        @get("/locktest")
        @render_json
        async def _locktest(_):
            secs = random.randint(1, 6)
            success = True

            try:
                async with mylock.lock("ponty"):  # "ponty" is the id of the protected resource
                    await asyncio.sleep(secs)
            except Locked:
                success = False

            return {
                "success": success,
                "secs": secs,
            }


    The five requests below are issued in parallel, so attempt to grab the lock
    at approximately the same time. Note how the first three acquire it at
    :math:`t = 0`, :math:`t = 1`, and :math:`t = 4`, respectively -  all
    beneath the 5-second timeout - while the last two time out at roughly the
    5-second mark. (With a larger timeout, they would have captured the lock
    at approx :math:`t = 8` and :math:`t = 10`.)

    .. code-block:: bash

        $ for i in {1..5}; do curl localhost:8080/locktest & done
        {"now": 1661271438402, "data": {"success": true, "secs": 1}, "elapsed": 1006}
        {"now": 1661271438402, "data": {"success": true, "secs": 3}, "elapsed": 4049}
        {"now": 1661271438402, "data": {"success": true, "secs": 4}, "elapsed": 8087}
        {"now": 1661271438403, "data": {"success": false, "secs": 2}, "elapsed": 5232}
        {"now": 1661271438404, "data": {"success": false, "secs": 5}, "elapsed": 5232}


    """
    return Lock(sentinels=_Sentinels(), **kw)



__all__ = ["dataclass_to_jsonschema", "Annotation"]


import dataclasses
from dataclasses import MISSING
import datetime
import enum
import numbers
import sys
import types
import typing

from ponty.types_ import DataclassProtocol


class Annotation(dict[str, typing.Any]):
    """
    Enrich member variable annotations with context-specific metadata.
    These annotations appear in the jsonschema as generic keywords;
    in some cases they are used for additional validation
    (e.g. `array.minItems`), while in others they simply help make the
    schema self-documenting (e.g. `description`).

    """




D = typing.TypeVar("D", bound=DataclassProtocol)


def dataclass_to_jsonschema(cls: type[D]) -> dict[str, typing.Any]:
    """Generate the jsonschema for a dataclass.

    Optionally, fields can be enriched by passing jsonschema keywords to
    :class:`Annotation`.

    Additionally, the dataclass can be enriched with keywords by embedding
    a JsonSchema class (using :class:`Annotation` as well).

    :param dataclass cls: dataclass definition
    :raises ValueError: raised if `cls` is not a dataclass
    :return: the JSON schema


    .. code-block:: python

        from dataclasses import dataclass
        import json
        import typing

        from ponty import Annotation, dataclass_to_jsonschema


        @dataclass
        class Model:

            name: str
            body_style: typing.Literal["coupe", "sedan", "hatchback", "convertible", "wagon", "suv"]
            year: typing.Annotated[int, Annotation(minimum=1886, maximum=2030)]
            msrp: typing.Annotated[int, Annotation(description="MSRP, dollars.")]
            cylinders: int = 4


        @dataclass
        class Make:

            name: typing.Annotated[str, Annotation(description="Brand name")]
            models: list[Model]

            class JsonSchema:
                annotation = Annotation(description="Automotive brand", examples=["Toyota", "Honda"])


        if __name__ == "__main__":
            schema = dataclass_to_jsonschema(Make)
            print(json.dumps(schema, indent=2))


    .. code-block:: json
        :caption: generated jsonschema

        {
          "$schema": "http://json-schema.org/draft-07/schema#",
          "properties": {
            "name": {
              "type": "string",
              "description": "Brand name"
            },
            "models": {
              "type": "array",
              "items": {
                "allOf": [
                  {
                    "$ref": "#/$defs/Model"
                  }
                ]
              }
            }
          },
          "title": "Make",
          "type": "object",
          "description": "Automotive brand",
          "examples": [
            "Toyota",
            "Honda"
          ],
          "required": [
            "name",
            "models"
          ],
          "$defs": {
            "Model": {
              "properties": {
                "name": {
                  "type": "string"
                },
                "body_style": {
                  "enum": [
                    "coupe",
                    "sedan",
                    "hatchback",
                    "convertible",
                    "wagon",
                    "suv"
                  ]
                },
                "year": {
                  "type": "integer",
                  "minimum": 1886,
                  "maximum": 2030
                },
                "msrp": {
                  "type": "integer",
                  "description": "MSRP, dollars."
                },
                "cylinders": {
                  "type": "integer",
                  "default": 4
                }
              },
              "title": "Model",
              "type": "object",
              "required": [
                "name",
                "body_style",
                "year",
                "msrp"
              ]
            }
          }
        }


    """
    if not dataclasses.is_dataclass(cls):
        raise ValueError("Dataclasses only.")

    defs: dict[str, typing.Any] = {}
    schema = _cache_dataclass(cls, cls, defs, True)

    if defs:
        schema["$defs"] = defs

    return {
        "$schema": "http://json-schema.org/draft-07/schema#",
        **schema,
    }




def _cache_dataclass(cls, primary, defs, firstpass=False):
    if cls is primary:
        if firstpass:
            return _render_dataclass(cls, primary, defs)
        return {"allOf": [{"$ref": "#"}]}

    name = cls.__name__
    if name not in defs:
        defs[name] = _render_dataclass(cls, primary, defs)

    return {"allOf": [{"$ref": f"#/$defs/{name}"}]}


def _render_dataclass(cls, primary, defs):
    properties = {}
    required = []

    for field in dataclasses.fields(cls):
        schema = _lookup(field.type, primary, defs)
        if field.default is not MISSING:
            schema["default"] = field.default

        properties[field.name] = schema

        if field.default is MISSING and field.default_factory is MISSING:
            required.append(field.name)

    schema = {
        "properties": properties,
        "title": cls.__name__,
        "type": "object",
        **getattr(getattr(cls, "JsonSchema", {}), "annotation", {})
    }

    if required:
        schema["required"] = required

    return schema


def _lookup(field_type, primary, defs, **kw):
    if dataclasses.is_dataclass(field_type):
        return _cache_dataclass(field_type, primary, defs)

    if field_type is typing.Any:
        return {}

    if field_type is None or field_type is type(None):
        return {"type": "null", **kw}

    unscripted_type = typing.get_origin(field_type)
    args = typing.get_args(field_type)

    if unscripted_type is typing.Annotated:
        for anno in args[1:]:
            if isinstance(anno, Annotation):
                break
        else:
            anno = {}

        return _lookup(args[0], primary, defs, **anno)

    if unscripted_type is typing.Literal:
        return {"enum": list(args), **kw}

    is_uniontype: bool = unscripted_type is typing.Union
    if sys.version_info >= (3, 10):
        is_uniontype = is_uniontype or (unscripted_type is types.UnionType)
    if is_uniontype:  # also catches Optional
        return {
            "anyOf": [_lookup(arg, primary, defs) for arg in args],
            **kw
        }

    if field_type is bool:
        return {"type": "boolean", **kw}
    if field_type is int:
        return {"type": "integer", **kw}
    if field_type is str:
        return {"type": "string", **kw}

    if field_type is dict or unscripted_type is dict:
        schema = {"type": "object", **kw}
        if len(args) == 2:
            t1, t2 = args
            if t1 is not str:
                raise Exception
            schema["additionalProperties"] = _lookup(t2, primary, defs)
        return schema

    if field_type is list or unscripted_type is list:
        schema = {"type": "array", **kw}
        if len(args) == 1:
            schema["items"] = _lookup(args[0], primary, defs)
        return schema

    if field_type is set or unscripted_type is set:
        schema = {"type": "array", "uniqueItems": True, **kw}
        if args:
            schema["items"] = _lookup(args[0], primary, defs)
        return schema

    if field_type is tuple or unscripted_type is tuple:
        schema = {"type": "array", **kw}

        if args:
            if len(args) == 2 and (args[1] is ...):
                schema["items"] = _lookup(args[0], primary, defs)
            else:
                schema["prefixItems"] = [_lookup(arg, primary, defs) for arg in args]
                schema["minItems"] = schema["maxItems"] = len(args)

        return schema

    if issubclass(field_type, datetime.datetime):  # subclass of datetime.date, so must come first
        return {"type": "string", "format": "date-time", **kw}
    if issubclass(field_type, datetime.date):
        return {"type": "string", "format": "date", **kw}
    if issubclass(field_type, numbers.Number):
        return {"type": "number", **kw}

    if issubclass(field_type, enum.Enum):
        name = field_type.__name__
        if name not in defs:
            defs[name] = {
                "title": name,
                "enum": [i.value for i in field_type],
                **kw
            }
        return {"allOf": [{"$ref": f"#/$defs/{name}"}]}

    raise Exception(f"no support for {field_type}/{unscripted_type}")


if __name__ == "__main__":
    import enum
    import json

    class Choices(enum.Enum):
        a = 1
        b = 2

    @dataclasses.dataclass
    class Bar:
        x: int
        y: int

    @dataclasses.dataclass
    class Foo:
        name: typing.Annotated[str, Annotation(minLength=22, pattern=".*")]
        age: int
        blah: list[int]
        blargh: dict[str, Bar]
        blam: tuple[Bar, ...]
        favorite_color: typing.Optional[str]
        enum_test: Choices
        union_test: typing.Union[str, int] = 7

        class JsonSchema:
            annotation = Annotation(description="go you!")

    schema = dataclass_to_jsonschema(Foo)
    print(json.dumps(schema, indent=2))



import typing

import aiohttp.web


class QueryParameter:
    """Extracts the value of query parameter with name `key` from the URI.

    :param str key: if provided, query parameter key name.
      By default, uses descriptor `__set_name__` to fetch the variable name

    :param bool required:
      if True, throws a 400 if the query parameter is not provided

    :param str default:
      default value, if the query param is not provided


    .. code-block:: python
        :emphasize-lines: 12

        from ponty import (
            expect,
            get,
            render_json,
            QueryParameter,
            Request,
        )


        class HelloReq(Request):

            punc = QueryParameter(default="!")


        @get("/hello")
        @expect(HelloReq)
        @render_json
        async def greet(punc: str):
            return {"greeting": f"hello world{punc}"}


    .. code-block:: bash
        :caption: the default in action
        :emphasize-lines: 4

        $ curl localhost:8080/hello | python -m json.tool
        {
            "data": {
                "greeting": "hello world!"
            },
            "elapsed": 0,
            "now": 1660439305592
        }


    .. code-block:: bash
        :caption: overriding the default
        :emphasize-lines: 4

        $ curl localhost:8080/hello?punc=. | python -m json.tool
        {
            "data": {
                "greeting": "hello world."
            },
            "elapsed": 0,
            "now": 1660439305592
        }

    """
    def __init__(
        self,
        *,
        key: str = None,
        required: bool = False,
        default: str = "",
    ):
        self._key = key
        self._required = required
        self._default = default

    def __set_name__(self, obj, name):
        if not self._key:
            self._key = name

    def __get__(self, obj, objtype=None) -> str:
        try:
            return obj.req.query[self._key]
        except KeyError:
            if self._required:
                raise aiohttp.web.HTTPBadRequest
            return self._default


class QueryParameterEnum(QueryParameter):
    """
    Inherits :class:`QueryParameter`.
    In addition to :class:`QueryParameter` behavior,
    validates the query parameter against a set of legal values.
    Throws a 400 in the event of a mismatch.

    :param values: legal values for the query parameter
    :param kw: additional parameters forwarded to :class:`QueryParameter`


    .. code-block:: python
        :emphasize-lines: 22

        from ponty import (
            expect,
            get,
            render_json,
            QueryParameterEnum,
            Request,
            StringRouteParameter,
        )


        _hellos: dict[str, str] = {
            "en": "hello",
            "es": "hola",
            "no": "hallo",
        }


        class HelloReq(Request):

            name = StringRouteParameter()
            lang = QueryParameterEnum(
                values=_hellos.keys(),
                default="en",
            )


        @get(f"/hello/{HelloReq.name}")
        @expect(HelloReq)
        @render_json
        async def ahoy(name: str, lang: str):
            # no need to trap the KeyError here, the descriptor has already
            # validated `lang` is a key in `_hellos`
            hi = _hellos[lang]
            return {"greeting": f"{hi} {name}!"}


    .. code-block:: bash

        $ curl "localhost:8080/hello/muchacho?lang=es" | python -m json.tool
        {
            "data": {
                "greeting": "hola muchacho!"
            },
            "elapsed": 0,
            "now": 1660440618107
        }

    """
    def __init__(self, *, values: typing.Iterable[str], **kw):
        super().__init__(**kw)
        self._values = values

        if self._default and self._default not in values:
            raise ValueError(f"default '{self._default}' does not validate")

    def __get__(self, obj, objtype=None) -> str:
        val = super().__get__(obj, objtype)
        if val not in self._values:
            msg = f"{self._key} must be one of {{{','.join(self._values)}}}"
            raise aiohttp.web.HTTPBadRequest(text=msg)
        return val



from setuptools import find_packages, setup


if __name__ == "__main__":
    setup(
        packages=find_packages(),
        package_data={"ponty": ["py.typed"]},
        name="ponty",
        version="0.3.6",
        license="BSD",
        url="https://github.com/csira/ponty",
        description="Minimal async web framework, built on aiohttp.",
        install_requires=[
            "aiohttp>=3.7.3",
            "jsonschema==3.2.0",
            "typing-extensions>=4.2.0",
        ],
        python_requires=">=3.8",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Web Environment",
            "Framework :: AsyncIO",
            "Framework :: aiohttp",
            "Intended Audience :: Developers",
            "License :: Freely Distributable",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Topic :: Internet :: WWW/HTTP",
            "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Typing :: Typed",
        ]
    )



__all__ = [
    "startmeup", "Provider",
    "get", "post", "put", "delete", "patch", "route", "route_iter",
    "expect", "Request",
    "RouteParameter", "PosIntRouteParameter", "StringRouteParameter",
    "TextBody", "JsonBody", "ValidatedJsonBody", "ParsedJsonBody",
    "QueryParameter", "QueryParameterEnum",
    "Header", "ContentLength", "ContentType",
    "Cookie", "AIOHttpReq",
    "render_json",
    "http_client_provider", "lease_http_client",
    "Annotation", "dataclass_to_jsonschema",

    "raise_status", "error_trap",
    "PontyError", "DoesNotExist", "ValidationError",

    "retry",
]


from ponty.errors import (
    raise_status, error_trap,
    PontyError, DoesNotExist, ValidationError,
)

from ponty.http import (
    startmeup, Provider,
    get, post, put, delete, patch, route, route_iter,
    expect, Request,
    RouteParameter, PosIntRouteParameter, StringRouteParameter,
    TextBody, JsonBody, ValidatedJsonBody, ParsedJsonBody,
    Header, ContentLength, ContentType,
    QueryParameter, QueryParameterEnum,
    Cookie, AIOHttpReq,
    render_json,
    http_client_provider, lease_http_client,
    Annotation, dataclass_to_jsonschema,
)

from ponty.utils import retry



import functools
import inspect
import typing
import warnings

import aiohttp.web


class JsonBody:
    """Contains the deserialized request body."""

    def __get__(self, obj: "Request", objtype: type["Request"]):
        return obj._json


class TextBody:
    """Contains the raw request body."""

    def __get__(self, obj: "Request", objtype: type["Request"]):
        return obj._text




T = typing.TypeVar("T", covariant=True)


@typing.runtime_checkable
class _Descriptor(typing.Protocol[T]):
    def __get__(self, obj, objtype=None) -> T: ...




class _Base(type):

    def __new__(cls, name, bases, dct):
        the_class = super().__new__(cls, name, bases, dct)

        fieldnames = []
        for key, val in dct.items():
            if isinstance(val, _Descriptor):
                fieldnames.append(key)
            if isinstance(val, JsonBody):
                the_class._extract_json = True
            if isinstance(val, TextBody):
                the_class._extract_text = True

        the_class._fieldnames = tuple(fieldnames)
        return the_class


class Request(metaclass=_Base):
    """
    Base class for request pre-processing.
    Instantiated by :func:`expect`.

    :ref:`Descriptors` for extracting components of the request should be
    stored as class variables on subclasses.

    :param req: automatically supplied by :func:`expect`
    :type req: `aiohttp.web.Request <https://docs.aiohttp.org/en/latest/web_reference.html#aiohttp.web.Request>`_

    """
    _fieldnames: tuple[str, ...] = ()
    _extract_json: bool = False
    _extract_text: bool = False

    def __init__(self, req: aiohttp.web.Request):
        self.req = req
        self._json = None
        self._text: typing.Optional[str] = None

    @property
    def _fields(self) -> dict[str, typing.Any]:
        return {name: getattr(self, name) for name in self._fieldnames}

    async def _prepare(self) -> None:
        if self._extract_json:
            self._json = await self.req.json()

        if self._extract_text:
            self._text = await self.req.text()




def expect(cls: type[Request], *, mimetype: str = None):
    """
    Preprocess the HTTP request, according to the rules configured in the
    :class:`Request` subclass.

    :param type[Request] cls: subclass of :class:`Request`,
      with rules for processing the HTTP request attached as descriptors

    :param str mimetype: expected IANA media type.
      If provided and the specified type does not match the media type
      found in the content-type header, throws a 415

    """
    def wraps(f):
        sig = inspect.signature(f)
        argnames = sig.parameters.keys()

        symdiff = set(cls._fieldnames) ^ set(argnames)
        if symdiff:
            raise TypeError(f"Parameter mismatch in '{f.__name__}': {', '.join(symdiff)}")

        if cls._extract_json:
            if mimetype is not None and mimetype != "application/json":
                warnings.warn(f"handler '{f.__name__}' expects a json payload but mimetype '{mimetype}' is expected")

        @functools.wraps(f)
        async def wrapper(req: aiohttp.web.Request):
            if mimetype:
                if mimetype not in req.headers["content-type"]:
                    raise aiohttp.web.HTTPUnsupportedMediaType

            inst = cls(req)
            await inst._prepare()
            return await f(**inst._fields)

        return wrapper
    return wraps



import abc
import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
import typing

from ponty.errors import PontyError


class Locked(PontyError):
    """Returns a 409. Inherits :class:`PontyError <ponty.PontyError>`."""

    status_code = 409


class SentinelStore(abc.ABC):
    """Container for asset id's currently protected by a :class:`Lock`.

    Abstact Base Class.
    :class:`Lock` expects the three abstract methods below -
    :meth:`exists <ponty.memo.SentinelStore.exists>`,
    :meth:`add <ponty.memo.SentinelStore.add>`,
    :meth:`remove <ponty.memo.SentinelStore.remove>` -
    to return awaitables, whether they perform IO or not.
    (NB no context switch occurs if the method does not perform a
    blocking operation.) This constraint may be relieved in a future version.

    """

    @abc.abstractmethod
    async def exists(self, key: str) -> bool:
        """Check if an asset is *currently* locked.

        :param key: unique id for the asset
        :returns: True if the asset is locked, False otherwise

        """

    @abc.abstractmethod
    async def add(self, key: str) -> None:
        """Take the lock on the asset.

        :param key: unique id for the asset

        """

    @abc.abstractmethod
    async def remove(self, key: str) -> None:
        """Release the lock on the asset.

        :param key: unique id for the asset

        """


@dataclass(frozen=True)
class Lock:
    """Basic mechanic for locking / antistampede.

    Operates as a mandatory lock by default. Control blocking by specifying
    the `maxwait_ms` parameter.

    :param sentinels: set-like container, providing add/remove operations and
      existence checking. Must be an instance of :class:`SentinelStore`

    :param maxwait_ms: total milliseconds to wait, before raising
      an exception with the caller.
      Raises `timeout_error` when :math:`n * pulse\_ms > maxwait\_ms`

    :param pulse_ms: milliseconds between subsequent attempts
      to acquire the lock

    :param timeout_error: exception class raised when `maxwait_ms` exceeded.
      If provided, must be a subclass of :class:`Locked`
    :type timeout_error: type[Locked]

    """
    sentinels: SentinelStore
    maxwait_ms: int = 0
    pulse_ms: int = 100
    timeout_error: type[Locked] = Locked

    @asynccontextmanager
    async def lock(self, key: str) -> typing.AsyncIterator[None]:
        """
        Context manager. Holds the lock on `key` until the block exits, or
        waits to acquire it until `maxwait_ms` milliseconds have elapsed.

        :param key: unique identifier for the asset whose access requires
          synchronization

        """
        elapsed = 0
        is_locked = await self.sentinels.exists(key)

        while is_locked and elapsed < self.maxwait_ms:
            await asyncio.sleep(self.pulse_ms * 0.001)
            elapsed += self.pulse_ms
            is_locked = await self.sentinels.exists(key)

        if is_locked:
            raise self.timeout_error(text=key)

        try:
            await self.sentinels.add(key)
            yield
        finally:
            await self.sentinels.remove(key)
